#include "countdown.h"

countdown_t volatile count;

#ifdef __GNUC__
#    define inline __attribute__((always_inline))
#endif // GCC

static inline void down   (uint8_t, char  volatile *, char);
static inline void down16 (uint8_t, short volatile *);

void down (uint8_t i, char volatile * cd, char m1)
{
    // Komponenten (i stck) von hinten nach vorne durchlaufen
    // Zeiger cd steht anfangs hinter der letzten Komponente, m1 = -1
    do
    {
        // Zeiger erniedrigen, Wert lesen und 1 abziehen
        char val = *(--cd) + m1;

        // Falls der Wert nicht abgelaufen ist wird er (vermindert)
        // zurckgeschrieben, ansonsten bleibt er bei 0.
        if (val != m1)
            *cd = val;
    } while (--i);
}

#define ITEM_COUNT(A,B) sizeof (A) / sizeof (B)

#define DOWN(x)             \
    do {                    \
        if (!sizeof (x))    \
            return;         \
        down (ITEM_COUNT(x, char), (char*) & x + ITEM_COUNT(x, char), m1); \
    } while (0)

void down16 (uint8_t i, volatile short *  cd)
{
    do
    {
        short val = *(--cd) -1;

        if (val != -1)
            *cd = val;
    } while (--i);
}

#define DOWN16(x)        \
    do {                 \
        if (sizeof (x))  \
            down16 (ITEM_COUNT(x, short), (void *) (& x + 1)); \
    } while (0)

// Alle 10 Millisekunden aufrufen
void job_countdown (void)
{
    char m1 = -1;

    __asm volatile ("" : "+r" (m1));
#ifdef COUNTDOWN_US100    
    DOWN (count.us100);

    if (count.us100.timer_10ms)
        return;
    count.us100.timer_10ms = 100;
#endif // COUNTDOWN_US100    

    // Die 10ms-Werte von `count' runterzhlen (8-Bit Werte, falls vorhanden)
    DOWN   (count.ms10);

    // Die 10ms-Werte von `count' runterzhlen (16-Bit Werte, falls vorhanden)
    DOWN16 (count.ms10_16);

    // .ms10.timer_1s abgelaufen?
    if (count.ms10.timer_1s)
        return;

    // Ja. Eine neue Sekunde beginnt: Sekunden-Zhler neu aufziehen
    count.ms10.timer_1s = 100;

    // Die Sekunden-Werte von `count' runterzhlen
    DOWN (count.sec);

    // Falls bei den Minuten nicht zu tun ist sind wir fertig.
    if (!sizeof (count.min))
        return;

    // Minuten-Zhler abgelaufen?
    if (count.sec.timer_1m)
        return;

    // Minuten-Zhler neu aufziehen
    count.sec.timer_1m = 60;

    // Die Minuten-Werte von `count' runterzhlen
    DOWN (count.min);
}
