#ifndef _AVR_PORT_MACROS_H_
#define _AVR_PORT_MACROS_H_

#ifndef PORT_ID
#define PORT_ID(n) \
    (n ## _ID << 3)
#endif /* PORT_ID */

#ifdef USE_SERPA

extern unsigned char serpa[];
extern unsigned char serpa_read[];
extern const unsigned char SERPA_NUM;
//#define SERPA_NUM 4

//unsigned char serpa[SERPA_NUM];

#define IS_SERPA_PORT(pinid) \
    ((SERPA_ID <= (pinid) >> 3) /* && (SERPA_ID+SERPA_NUM > (pinid) >> 3)*/)

#define N_TO_SERPA(pinid) \
    serpa[((pinid) >> 3)-SERPA_ID]

#define N_TO_SERPAIN(pinid) \
    serpa_read[0]

#define IF_SERPA_OP_ELSE(pinid,op) \
    if (IS_SERPA_PORT (pinid))\
        N_TO_SERPA(pinid) op ID_TO_BIT(pinid); \
    else

#else  /* USE_SERPA */

/*#define SERPA_NUM 0*/
#define IF_SERPA_OP_ELSE(pinid,op)
#define N_TO_SERPA(pinid) 0
#define N_TO_SERPAIN(pinid) 0

#endif /* USE_SERPA */

/*
#define cbi(a,b) \
    __asm volatile ("cbi\t%0, %1" :: "M" (_SFR_IO_ADDR (a)), "M" (b))
*/    

#define IS_HARD_PORT(pinid) \
    ((pinid >> 3) < SERPAIN_ID)

/** Set a PORT bit to 1 */
#define SET(pinid) \
    do { \
        /*__asm volatile (""::);*/\
        if ((pinid) >> 3 == SERPAIN_ID); else \
        IF_SERPA_OP_ELSE(pinid, |=) \
        N_TO_PORT (pinid) |= ID_TO_BIT(pinid); \
    } while (0)

#define BARRIER_CSE __asm volatile (""::)

/** Set a PORT bit to 0 */
#define CLR(pinid) \
    do { \
        BARRIER_CSE;                           \
        if ((pinid) >> 3 == SERPAIN_ID); else  \
        IF_SERPA_OP_ELSE(pinid, &= ~)          \
        N_TO_PORT (pinid) &= ~ID_TO_BIT(pinid);\
        BARRIER_CSE;                           \
    } while (0)

/** Toggle a PORT bit.
 * Note: The generated code will NOT be atomic!
 */
#define TOGGLE(pinid)                           \
    do {                                        \
        BARRIER_CSE;                           \
        if ((pinid) >> 3 == SERPAIN_ID); else   \
        IF_SERPA_OP_ELSE(pinid, ^= )            \
        N_TO_PORT (pinid) ^= ID_TO_BIT(pinid);  \
        /*N_TO_PIN (pinid) ^= ID_TO_BIT(pinid);*/   \
        BARRIER_CSE;                           \
    } while (0)

/** Set DDR to 1 */
#define MAKE_OUT(pinid)                             \
    do {                                            \
        /*__asm volatile (""::);                   */     \
        if (IS_HARD_PORT(pinid))                    \
            N_TO_DDR  (pinid) |= ID_TO_BIT(pinid);  \
    } while (0)

/** Set DDR to 0 */
#define MAKE_IN(pinid)                              \
    do {                                            \
        __asm volatile (""::);                        \
        if (IS_HARD_PORT(pinid))                    \
            N_TO_DDR  (pinid) &= ~ID_TO_BIT(pinid); \
        __asm volatile (""::);                        \
    } while (0)

/** Is PIN == 1 ? */
#define IS_SET(pinid) \
    (IS_HARD_PORT(pinid) \
        ? (N_TO_PIN (pinid) & ID_TO_BIT(pinid)) \
        : (unsigned char) ((unsigned char) N_TO_SERPAIN(pinid) & (unsigned char)    ID_TO_BIT(pinid)))

/** Is PIN == 0 ? */
#define IS_CLEAR(pinid) \
    (!IS_SET(pinid))

/** Is PORT == 1 ? */
#define IS_PORT(pinid)  (N_TO_PORT (pinid) & ID_TO_BIT(pinid))

/*
 * IDs for the various PORTs of our AVRs.
 * If you want to make this work for targets with
 * even more ports (like ATMega128 etc) just add these
 * ports in the following macro definitions.
 * Using #error would require making the macros target dependent.
 */
enum {
    PORTA_ID,
    PORTB_ID,
    PORTC_ID,
    PORTD_ID,

    SERPAIN_ID,
    SERPA_ID
};

#define pDDRX(pinid)  ({__asm volatile (".bad_ddr_id_" # pinid);(unsigned char*) 0;})
#define pPORTX(pinid) ({__asm volatile (".bad_port_id_" # pinid);(unsigned char*) 0;})
#define pPINX(pinid)  ({__asm volatile (".bad_pin_id_" # pinid);(unsigned char*) 0;})

/*
 * #define addresses of PORTs or an assembler error if
 * a non-existing PORT is refered.
 * I found no way to tell this error by the (pre)compiler.
 * Note that #error does not work as expected here.
 */
#if defined(PORTA)
#define pPORTA &PORTA
#else
#define pPORTA ({__asm volatile (".target_has_no_porta");(unsigned char*) 0;})
#endif

#if defined(PORTB)
#define pPORTB &PORTB
#else
#define pPORTB ({__asm volatile (".target_has_no_portb");(unsigned char*) 0;})
#endif

#if defined(PORTC)
#define pPORTC &PORTC
#else
#define pPORTC ({__asm volatile (".target_has_no_portc");(unsigned char*) 0;})
#endif

#if defined(PORTD)
#define pPORTD &PORTD
#else
#define pPORTD ({__asm volatile (".target_has_no_portd");(unsigned char*) 0;})
#endif

/*
 * #define addresses of DDRs or an assembler error if
 * a non-existing DDR is refered.
 */
#if defined(DDRA)
#define pDDRA &DDRA
#else
#define pDDRA ({__asm volatile (".target_has_no_ddra");(unsigned char*) 0;})
#endif

#if defined(DDRB)
#define pDDRB &DDRB
#else
#define pDDRB ({__asm volatile (".target_has_no_ddrb");(unsigned char*) 0;})
#endif

#if defined(DDRC)
#define pDDRC &DDRC
#else
#define pDDRC ({__asm volatile (".target_has_no_ddrc");(unsigned char*) 0;})
#endif

#if defined(DDRD)
#define pDDRD &DDRD
#else
#define pDDRD ({__asm volatile (".target_has_no_ddrd");(unsigned char*) 0;})
#endif

/*
 * #define addresses of PINs or an assembler error if
 * a non-existing PIN is refered.
 */
#if defined(PINA)
#define pPINA &PINA
#else
#define pPINA ({__asm volatile (".target_has_no_pina");(unsigned char*) 0;})
#endif

#if defined(PINB)
#define pPINB &PINB
#else
#define pPINB ({__asm volatile (".target_has_no_pinb");(unsigned char*) 0;})
#endif

#if defined(PINC)
#define pPINC &PINC
#else
#define pPINC ({__asm volatile (".target_has_no_pinc");(unsigned char*) 0;})
#endif

#if defined(PIND)
#define pPIND &PIND
#else
#define pPIND ({__asm volatile (".target_has_no_pind");(unsigned char*) 0;})
#endif


#define N_TO_pPORT(pinid)                       \
    (                                           \
        (PORTA_ID == (pinid) >> 3) ? pPORTA :   \
        (PORTB_ID == (pinid) >> 3) ? pPORTB :   \
        (PORTC_ID == (pinid) >> 3) ? pPORTC :   \
        (PORTD_ID == (pinid) >> 3) ? pPORTD :   \
        pPORTX(pinid)                           \
    )

#define N_TO_pDDR(pinid)                        \
    (                                           \
        (PORTA_ID == (pinid) >> 3) ? pDDRA :    \
        (PORTB_ID == (pinid) >> 3) ? pDDRB :    \
        (PORTC_ID == (pinid) >> 3) ? pDDRC :    \
        (PORTD_ID == (pinid) >> 3) ? pDDRD :    \
        pDDRX(pinid)                            \
    )

#define N_TO_pPIN(pinid)                        \
    (                                           \
        (PORTA_ID == (pinid) >> 3) ? pPINA :    \
        (PORTB_ID == (pinid) >> 3) ? pPINB :    \
        (PORTC_ID == (pinid) >> 3) ? pPINC :    \
        (PORTD_ID == (pinid) >> 3) ? pPIND :    \
        pPINX(pinid)                            \
    )

#define N_TO_PORT(pinid) (*(N_TO_pPORT(pinid)))
#define N_TO_DDR(pinid)  (*(N_TO_pDDR(pinid)))
#define N_TO_PIN(pinid)  (*(N_TO_pPIN(pinid)))

#define CHECK_PORT_MACRO(port) \
do { \
    if (&PORT ## port != N_TO_pPORT (PIN_ID (port,0))) __asm volatile (".error_in_N_TO_pPORT_" #port); \
    if (&DDR  ## port != N_TO_pDDR  (PIN_ID (port,0))) __asm volatile (".error_in_N_TO_pDDR_" #port); \
    if (&PIN  ## port != N_TO_pPIN  (PIN_ID (port,0))) __asm volatile (".error_in_N_TO_pPIN_" #port); \
} while (0)

#endif /* _AVR_PORT_MACROS_H_ */
